from jax.numpy import split as split_function
from .explode import explode


def split(pytree):
    return explode(
        lambda array: split_function(array, array.shape[-1], axis=-1), pytree
    )



from abc import ABC, abstractmethod
from .box import Box


class Model(Box, ABC):
    def __new__(cls, *args, **kwargs):
        if len(args) == 1 and not kwargs and isinstance(args[0], Box):
            (box,) = args
            (pytree,) = box
        else:
            pytree = cls.constructor(*args, **kwargs)
        return super().__new__(cls, pytree)

    @classmethod
    def tree_unflatten(cls, _, children):
        (pytree,) = children
        return cls(Box(pytree))

    def __call__(self, *args, **kwargs):
        (pytree,) = self
        return self.__class__.apply(pytree, *args, **kwargs)

    @staticmethod
    @abstractmethod
    def constructor(*args, **kwargs):
        pass

    @staticmethod
    @abstractmethod
    def apply(*args, **kwargs):
        pass



from functools import partial
from jax.tree_util import tree_map
from jax import grad, jit
from .loss import loss
from .gd import gd


@jit
def fit(model, x, y, learning_rate=0.1):
    gradients = grad(loss)(model, x, y)
    return tree_map(partial(gd, learning_rate), model, gradients)



from jax.tree_util import register_pytree_node_class


@register_pytree_node_class
class Box(tuple):
    def __new__(cls, x):
        return super().__new__(cls, (x,))

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        register_pytree_node_class(cls)

    def tree_flatten(self):
        (x,) = self
        return (x,), None

    @classmethod
    def tree_unflatten(cls, _, children):
        (x,) = children
        return cls(x)



from functools import partial
from contextlib import suppress
from jax.tree_util import tree_flatten, tree_unflatten


def explode(function, pytree):
    leaves, treedef = tree_flatten(pytree)
    entries = list(map(function, leaves))
    entries = list(consume(entries))
    return list(map(partial(tree_unflatten, treedef), entries))


def shift(a):
    return a.pop(0)


def consume(entries):
    with suppress(IndexError):
        while True:
            yield list(map(shift, entries))



from .integer import integer
from .weight import weight
from .param import param
from .bias import bias
from .seed import seed
from .key import key

__all__ = [
    "integer",
    "weight",
    "param",
    "bias",
    "seed",
    "key",
]



from collections.abc import Callable
from multipledispatch import dispatch
from .identity import identity
from .model import Model
from .box import Box


@dispatch(str, Callable, Callable)
def bind(name, constructor, apply):
    return type(name, (Model,), {"constructor": constructor, "apply": apply})


@dispatch(Callable, Callable)
def bind(constructor, apply):
    return bind(Model.__name__, constructor, apply)


@dispatch(object, Callable)
def bind(x, apply):
    cls = bind(identity, apply)
    return cls(Box(x))



from . import random
from . import vjp
from .apply import apply
from .binary import binary
from .bind import bind
from .box import Box
from .brelu1 import brelu1
from .compose import compose
from .concat import concat
from .explode import explode
from .fit import fit
from .gd import gd
from .identity import identity
from .implode import implode
from .loss import loss
from .model import Model
from .mse import mse
from .neuron import neuron
from .relu import relu
from .relu1 import relu1
from .spark import spark
from .spike import spike
from .split import split
from .stack import stack
from .unstack import unstack

Neuron = bind("Neuron", neuron, apply)
Binary = compose("Binary", Neuron, binary)
BReLU1 = compose("BReLU1", Neuron, brelu1)
ReLU = compose("ReLU", Neuron, relu)
ReLU1 = compose("ReLU1", Neuron, relu1)
Spark = BReLU1
Spike = Binary

__all__ = [
    "BReLU1",
    "Binary",
    "Box",
    "Model",
    "Neuron",
    "ReLU",
    "ReLU1",
    "Spark",
    "Spike",
    "apply",
    "binary",
    "bind",
    "brelu1",
    "compose",
    "concat",
    "explode",
    "fit",
    "gd",
    "identity",
    "implode",
    "loss",
    "mse",
    "neuron",
    "random",
    "relu",
    "relu1",
    "spark",
    "spike",
    "split",
    "stack",
    "unstack",
    "vjp",
]



from jax.numpy import ndarray
from multipledispatch import dispatch
from .random.key import key as random_key
from .random.weight import weight
from .random.bias import bias


@dispatch(ndarray, int)
def neuron(key, n):
    return (weight(key, (n, 1)), bias(key, (1, 1)))


@dispatch(int)
def neuron(n):
    return neuron(random_key(), n)



from jax import jit


def compose(name, cls, function):
    function = jit(function)

    def __call__(self, x):
        return function(cls.__call__(self, x))

    return type(name, (cls,), {"__call__": __call__})



from jax import custom_vjp


def identity(function):
    wrapper = custom_vjp(function)

    def forward(x):
        return function(x), None

    def backward(_, gradient):
        return (gradient,)

    wrapper.defvjp(forward, backward)
    return wrapper
